'''
server
    Created on 21 Nov 2018
@author: Tom Knot <knot@faster.cz>
'''
from pymodbus.server.sync import StartTcpServer
from pymodbus.device import ModbusDeviceIdentification
from pymodbus.datastore import ModbusSparseDataBlock
from pymodbus.datastore import ModbusSlaveContext, ModbusServerContext
from lightbus import UniPiLightBus, COMMAND_LIST
from time import sleep
import unipi_modbus_client
import threading
import logging

logging.basicConfig()
log = logging.getLogger()
log.setLevel(logging.ERROR)

local_modbus_client = unipi_modbus_client.UniPiModbusClient()
light_busses = []
sleeper_wakeup_thread = None
historical_coil_values = {}
historical_register_values = {}

class RegisterDataBlock(ModbusSparseDataBlock):
    def __init__(self):
        super(RegisterDataBlock, self).__init__({0:0})

    def setValues(self, address, value):
        for x in range(0, len(value)):
            if (not(address + x in historical_register_values.keys())) or historical_register_values[address + x] != value[x]:
                if (address + x) < 50:
                    physical_address = (address + x) % 10
                    if physical_address == 1:
                        try:
                            light_busses[0].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 2):
                        try:
                            light_busses[0].do_work([light_busses[0].perform_duplicate_write, [14, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 3):
                        try:
                            light_busses[1].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)                   
                    elif (physical_address == 4):
                        try:
                            light_busses[1].do_work([light_busses[1].perform_duplicate_write, [16, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 5):
                        try:
                            light_busses[2].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 6):
                        try:
                            light_busses[2].do_work([light_busses[2].perform_duplicate_write, [18, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 7):
                        try:
                            light_busses[3].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 8):
                        try:
                            light_busses[3].do_work([light_busses[3].perform_duplicate_write, [20, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                elif (address + x) < 100:
                    physical_address = (address + x) % 10
                    if physical_address == 1:
                        try:
                            light_busses[0].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 2):
                        try:
                            light_busses[0].do_work([light_busses[0].perform_single_write, [14, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 3):
                        try:
                            light_busses[1].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)                   
                    elif (physical_address == 4):
                        try:
                            light_busses[1].do_work([light_busses[1].perform_single_write, [16, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 5):
                        try:
                            light_busses[2].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 6):
                        try:
                            light_busses[2].do_work([light_busses[2].perform_single_write, [18, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 7):
                        try:
                            light_busses[3].opt_value_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 8):
                        try:
                            light_busses[3].do_work([light_busses[3].perform_single_write, [20, value[x], address + x]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                elif address + x < 200:
                    physical_address = (address + x - 101) % 3                    
                    if (physical_address == 0):
                        try:
                            light_busses[0].tx_address_dict[address + x + 2] = light_busses[0].construct_address(value[x])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 1):
                        try:
                            light_busses[0].tx_argument_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 2):
                        try:
                            if value[x] in COMMAND_LIST:
                                cmnd = light_busses[0].get_command_name(value[x])
                                if light_busses[0].command_has_argument(value[x]):
                                    light_busses[0].do_work([light_busses[0].perform_command, [cmnd, light_busses[0].get_tx_address(address + x), light_busses[0].get_tx_argument(address + x)]])
                                else:
                                    light_busses[0].do_work([light_busses[0].perform_command, [cmnd, light_busses[0].get_tx_address(address + x), None]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                elif address + x < 300:
                    physical_address = (address + x - 201) % 3              
                    if (physical_address == 0):
                        try:
                            light_busses[1].tx_address_dict[address + x + 2] = light_busses[1].construct_address(value[x])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 1):
                        try:
                            light_busses[1].tx_argument_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 2):
                        try:
                            if value[x] in COMMAND_LIST:
                                cmnd = light_busses[1].get_command_name(value[x])
                                if light_busses[1].command_has_argument(value[x]):
                                    light_busses[1].do_work([light_busses[1].perform_command, [cmnd, light_busses[1].get_tx_address(address + x), light_busses[1].get_tx_argument(address + x)]])
                                else:
                                    light_busses[1].do_work([light_busses[1].perform_command, [cmnd, light_busses[1].get_tx_address(address + x), None]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                elif address + x < 400:
                    physical_address = (address + x - 301) % 3              
                    if (physical_address == 0):
                        try:
                            light_busses[2].tx_address_dict[address + x + 2] = light_busses[2].construct_address(value[x])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 1):
                        try:
                            light_busses[2].tx_argument_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 2):
                        try:
                            if value[x] in COMMAND_LIST:
                                cmnd = light_busses[2].get_command_name(value[x])
                                if light_busses[2].command_has_argument(value[x]):
                                    light_busses[2].do_work([light_busses[2].perform_command, [cmnd, light_busses[2].get_tx_address(address + x), light_busses[2].get_tx_argument(address + x)]])
                                else:
                                    light_busses[2].do_work([light_busses[2].perform_command, [cmnd, light_busses[2].get_tx_address(address + x), None]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                elif address + x < 500:
                    physical_address = (address + x - 401) % 3              
                    if (physical_address == 0):
                        try:
                            light_busses[3].tx_address_dict[address + x + 2] = light_busses[3].construct_address(value[x])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 1):
                        try:
                            light_busses[3].tx_argument_dict[address + x + 1] = value[x]
                        except BaseException, E:
                            print "Exc: " + str(E)
                    elif (physical_address == 2):
                        try:
                            if value[x] in COMMAND_LIST:
                                cmnd = light_busses[3].get_command_name(value[x])
                                if light_busses[3].command_has_argument(value[x]):
                                    light_busses[3].do_work([light_busses[3].perform_command, [cmnd, light_busses[3].get_tx_address(address + x), light_busses[3].get_tx_argument(address + x)]])
                                else:
                                    light_busses[3].do_work([light_busses[3].perform_command, [cmnd, light_busses[3].get_tx_address(address + x), None]])
                        except BaseException, E:
                            print "Exc: " + str(E)
                historical_register_values[address + x] = value[x]
    def getValue(self, address):
        ret = 0
        if address in historical_register_values.keys():
            ret = historical_register_values[address]
        return [ret]
                
    def getValues(self, address, count=1):
        ret = []
        for x in range(0, count):
            ret += self.getValue(address + x)
        return ret
    
    def validate(self, address, count=1):
        if count == 0: return False
        max_value = address + count
        if max_value <= 500:
            return True
        return False

class CoilDataBlock(ModbusSparseDataBlock):
    def __init__(self):
        super(CoilDataBlock, self).__init__({0:0})

    def setValues(self, address, value):
        global local_serial_server
        for x in range(0, len(value)):
            if (not(address + x in historical_coil_values.keys())) or historical_coil_values[address + x] != value[x]:
                if (address + x == 1):
                    if value[x]:
                        try:
                            light_busses[0].do_work([light_busses[0].scan, []])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    historical_coil_values[address + x] = value[x]
                elif (address + x == 2):
                    if value[x]:
                        try:
                            light_busses[1].do_work([light_busses[1].scan, []])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    historical_coil_values[address + x] = value[x]
                elif (address + x == 3):
                    if value[x]:
                        try:
                            light_busses[2].do_work([light_busses[2].scan, []])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    historical_coil_values[address + x] = value[x]
                elif (address + x == 4):
                    if value[x]:
                        try:
                            light_busses[3].do_work([light_busses[3].scan, []])
                        except BaseException, E:
                            print "Exc: " + str(E)
                    historical_coil_values[address + x] = value[x]
                
    def getValues(self, address, count=1):
        ret = []
        for x in range(0, count):
            ret += self.getValue(address + x)
        return ret
        
    def getValue(self, address):
        ret = 0
        if address in historical_coil_values.keys():
            ret = historical_coil_values[address]
        return [ret]
            
    def validate(self, address, count=1):
        if count == 0: return False
        max_value = address + count
        if max_value <= 5:
            return True
        return False

class UniPiDaliModbusServer(object):
    coil_data = None
    register_data = None
    data_store = None
    server_identity = None
    server_context = None
    server_thread = None
    
    def __init__(self):
        self.coil_data = CoilDataBlock()
        self.register_data = RegisterDataBlock()
        self.store = ModbusSlaveContext(
            co = self.coil_data,
            hr = self.register_data
        )
        self.server_identity = ModbusDeviceIdentification()
        self.server_identity.VendorName  = 'UniPi Technology'
        self.server_identity.ProductCode = 'SUPDMS'
        self.server_identity.VendorUrl   = 'http://www.unipi.technology/'
        self.server_identity.ProductName = 'Simple UniPi DALI Modbus Server'
        self.server_identity.ModelName   = 'Simple UniPi DALI Modbus Server'
        self.server_identity.MajorMinorRevision = '1.0'

            
    def run(self):
        self.server_context = ModbusServerContext(slaves=self.store, single=True)
        try:
            local_modbus_client.write_register(20, 0x00)
        except BaseException, E:
            print "Exc: " + str(E)
        StartTcpServer(self.server_context, identity=self.server_identity, address=('0.0.0.0', 504))   

def sleeper_wakeup_loop():
    while(True):
        for light_bus in light_busses:
            if (len(light_bus.tx_func_queue) > 0) and (not light_bus.is_thread_active()):
                light_bus.wakeup_thread()
        sleep(0.25)

if __name__ == '__main__':
    """
        Usage: python unipidali.py address value
    """
    server = UniPiDaliModbusServer()
    logger = logging.getLogger('UnipiDALIDriver')
    for i in range(4):
        light_busses.append(UniPiLightBus(local_modbus_client, bus=i))
        light_busses[i].driver.logger = logger
        light_busses[i].driver.debug = True
    sleeper_wakeup_thread = threading.Thread(target=sleeper_wakeup_loop, args=[])
    sleeper_wakeup_thread.start()
    server.run()
